import {Layer, Value, StreamChunk, Buffer, LargeBuffer} from 'dripcap';
import Enum from 'driptool/enum';

export default class Dissector {
  static get namespaces() {
    return [/::Ethernet::\w+::TCP/];
  }

  constructor() {
    this.request = 0;
    this.decoder = this.decode();
    this.buffer = Buffer.from([]);
    this.c2s = null;
  }

  analyze(packet, parentLayer, chunk) {
    if (!this.decoder) return;

    if (parentLayer.getValue('srcPort').data !== 25565 &&
        parentLayer.getValue('dstPort').data !== 25565) {
      return;
    }

    let payload = chunk.attrs.payload.data;
    if (payload.length === 0) return;
    this.buffer = Buffer.concat([this.buffer, payload]);

    let layers = [];
    while (this.request <= this.buffer.length) {
      let result = this.decoder.next(this.buffer.slice(0, this.request));
      if (this.request > 0) this.buffer = this.buffer.slice(this.request);
      if (result.done) break;
      if (result.value === null) {
        this.buffer = Buffer.from([]);
        this.request = 0;
        return;
      } else if (Number.isInteger(result.value)) {
        this.request = result.value;
      } else if (result.value) {

        let layer = {
          items: [],
          namespace: chunk.namespace + '::Minecraft',
          name: 'Minecraft',
          id: 'minecraft',
          payload: chunk.attrs.payload.data
        };

        Object.assign(layer, result.value);
        layer.items.push({id: 'src', value: parentLayer.getValue('src')});
        layer.items.push({id: 'dst', value: parentLayer.getValue('dst')});

        layers.push(new Layer(layer));
        this.request = 0;
      }
    }
    return layers;
  }

  *decodeMetadata() {
    while (true) {
      let index = (yield 1).readUInt8(0);
      if (index === 0xff) break;
      let type = (yield 1).readUInt8(0);
/*
      switch (type) {
        case 0:
          yield 1;
          break;
        case 1:
          do {
            let int = yield 1;
          } while (int[0] & 0b10000000);
          break;
        default:
          console.log('unknown entity type: ' + type);
          yield null;
          return;
      }
*/
    }
  }

  *decode() {
    while (true) {
      let idData = yield 1;
      let id = idData[0];
      switch (id) {
        case 0x00:
        {
          let numData = yield 4;
          let num = numData.readInt32BE(0);
          yield {
            payload: Buffer.concat([idData, numData]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'Keep-alive ID',
                id: 'keepAlive',
                range: '1:',
                value: num
              }
            ],
            summary: 'Keep Alive'
          };
        }
          break;
        case 0x01:
        {
          let versionData = yield 4;
          let version = versionData.readInt32BE(0);
          if (version !== 29) {
            yield null;
            continue;
          }

          let lenData = yield 2;
          let len = lenData.readUInt16BE(0);
          let nameData = (yield len * 2);
          let name = utf16(nameData);

          let levelLenData = yield 2;
          let levelLen = levelLenData.readUInt16BE(0);
          let levelData = (yield levelLen * 2);
          let level = utf16(levelData);

          let reservedData = yield 11;
          let payload = Buffer.concat([idData, versionData, lenData, nameData, levelLenData, levelData, reservedData]);

          if (this.c2s) {
            yield {
              payload: payload,
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'Protocol Version',
                  id: 'version',
                  range: '1:5',
                  value: version
                },
                {
                  name: 'Username',
                  id: 'name',
                  range: `5:${7+len*2}`,
                  value: name
                }
              ],
              summary: 'Login Request C->S'
            };
          } else {
            yield {
              payload: payload,
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'Entity ID',
                  id: 'entitiy',
                  range: '1:5',
                  value: version
                },
                {
                  name: 'Level Type',
                  id: 'levelType',
                  range: `${7+len*2}:${7+len*2+2+levelLen*2}`,
                  value: level
                }
              ],
              summary: 'Login Request S->C'
            };
          }
        }
          break;
        case 0x02:
        {
          let lenData = yield 2;
          let len = lenData.readUInt16BE(0);
          let nameData = (yield len * 2);
          let name = utf16(nameData);
          if (name.includes(';')) {
            this.c2s = true;
            yield {
              payload: Buffer.concat([idData, lenData, nameData]),
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'Username and Host',
                  id: 'uhost',
                  range: '2:',
                  value: name
                }
              ],
              summary: 'Handshake C->S'
            };
          } else {
            this.c2s = false;
            yield {
              payload: Buffer.concat([idData, lenData, nameData]),
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'Connection Hash',
                  id: 'hash',
                  range: '2:',
                  value: name
                }
              ],
              summary: 'Handshake S->C'
            };
          }
        }
          break;

        case 0x04:
        {
          let data = yield 8;
          let time = data.readUInt32BE(5);
          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'Time',
                id: 'time',
                range: '1:9',
                value: time
              },
              {
                name: ' ',
                value: new Value(time, 'minecraft-time')
              },
            ],
            summary: 'Time Update'
          };
        }
          break;

        case 0x06:
        {
          let data = yield 12;
          let x = data.readInt32BE(0);
          let y = data.readInt32BE(4);
          let z = data.readInt32BE(8);
          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'X',
                id: 'x',
                range: '1:5',
                value: x
              },
              {
                name: 'Y',
                id: 'y',
                range: '5:9',
                value: y
              },
              {
                name: 'Z',
                id: 'z',
                range: '9:13',
                value: z
              },
            ],
            summary: 'Spawn Position S->C'
          };
        }
          break;

        case 0x0a:
        {
          let data = yield 1;
          let ground = !!data.readUInt8(0);
          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'On Ground',
                id: 'ground',
                range: '1:2',
                value: ground
              },
            ],
            summary: 'Player'
          };
        }
          break;

        case 0x0b:
        {
          let data = yield 33;
          let x = data.readDoubleBE(0);
          let y = data.readDoubleBE(8);
          let s = data.readDoubleBE(16);
          let z = data.readDoubleBE(24);
          let ground = !!data.readUInt8(32);

          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'Position',
                id: 'pos',
                range: '1:33',
                value: `x: ${x} y: ${y} z: ${z} stance: ${s}`,
                items: [
                  {
                    name: ' ',
                    value: new Value([x, y, z], 'minecraft-pos')
                  },
                  {
                    name: 'X',
                    id: 'x',
                    range: '1:9',
                    value: x
                  },
                  {
                    name: 'Y',
                    id: 'y',
                    range: '9:17',
                    value: y
                  },
                  {
                    name: 'Stance',
                    id: 'stance',
                    range: '17:25',
                    value: s
                  },
                  {
                    name: 'Z',
                    id: 'z',
                    range: '25:33',
                    value: z
                  },
                ]
              },
              {
                name: 'On Ground',
                id: 'ground',
                range: '33:34',
                value: ground
              },
            ],
            summary: 'Player Position'
          };
        }
          break;

        case 0x0d:
        {
          let data = yield 41;
          let x = data.readDoubleBE(0);
          let y = data.readDoubleBE(8);
          let s = data.readDoubleBE(16);
          let z = data.readDoubleBE(24);
          let yaw = data.readFloatBE(32);
          let pitch = data.readFloatBE(36);
          let ground = !!data.readUInt8(40);

          if (this.c2s) {
            yield {
              payload: Buffer.concat([idData, data]),
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'Position',
                  id: 'pos',
                  range: '1:33',
                  value: `x: ${x} y: ${y} z: ${z} stance: ${s}`,
                  items: [
                    {
                      name: ' ',
                      value: new Value([x, y, z], 'minecraft-pos')
                    },
                    {
                      name: 'X',
                      id: 'x',
                      range: '1:9',
                      value: x
                    },
                    {
                      name: 'Y',
                      id: 'y',
                      range: '9:17',
                      value: y
                    },
                    {
                      name: 'Stance',
                      id: 'stance',
                      range: '17:25',
                      value: s
                    },
                    {
                      name: 'Z',
                      id: 'z',
                      range: '25:33',
                      value: z
                    },
                  ]
                },
                {
                  name: 'Yaw',
                  id: 'yaw',
                  range: '33:37',
                  value: yaw
                },
                {
                  name: 'Pitch',
                  id: 'pitch',
                  range: '37:41',
                  value: pitch
                },
                {
                  name: 'On Ground',
                  id: 'ground',
                  range: '41:42',
                  value: ground
                },
              ],
              summary: 'Player Position & Look C->S'
            };
          } else {
            yield {
              payload: Buffer.concat([idData, data]),
              items: [
                {
                  name: 'Packet ID',
                  id: 'id',
                  range: '0:1',
                  value: id
                },
                {
                  name: 'X',
                  id: 'x',
                  range: '1:9',
                  value: x
                },
                {
                  name: 'Stance',
                  id: 'stance',
                  range: '9:17',
                  value: y
                },
                {
                  name: 'Y',
                  id: 'y',
                  range: '17:25',
                  value: s
                },
                {
                  name: 'Z',
                  id: 'z',
                  range: '25:33',
                  value: z
                },
                {
                  name: 'Yaw',
                  id: 'yaw',
                  range: '33:37',
                  value: yaw
                },
                {
                  name: 'Pitch',
                  id: 'pitch',
                  range: '37:41',
                  value: pitch
                },
                {
                  name: 'On Ground',
                  id: 'ground',
                  range: '41:42',
                  value: ground
                },
              ],
              summary: 'Player Position & Look S->C'
            };
          }
        }
          break;

        case 0x15:
        {
          let data = yield 24;
          let eid = data.readUInt32BE(0);
          let item = data.readUInt16BE(4);
          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'EID',
                id: 'eid',
                range: '1:5',
                value: eid
              },
              {
                name: 'Item',
                id: 'item',
                range: '5:7',
                value: item
              },
            ],
            summary: 'Spawn Dropped Item'
          };
        }
          break;
          /*
        case 0x18:
        {
          let data = yield 20;
          let eid = data.readUInt32BE(0);
          let type = data.readUInt8(4);
          let x = data.readInt32BE(5);
          let y = data.readInt32BE(9);
          let z = data.readInt32BE(13);
          let yaw = data.readInt8(17);
          let pitch = data.readInt8(18);
          let headYaw = data.readInt8(19);

          let reader = this.decodeMetadata();
          for (let result = reader.next(); !result.done ;) {
            if (result.value === null) {
              yield null;
            }
            let buf = yield result.value;
            data = Buffer.concat([data, buf]);
            result = reader.next(buf);
          }

          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'EID',
                id: 'eid',
                range: '1:5',
                value: eid
              },
              {
                name: 'Type',
                id: 'type',
                range: '5:6',
                value: type
              },
              {
                name: 'X',
                id: 'x',
                range: '6:10',
                value: x
              },
              {
                name: 'Y',
                id: 'y',
                range: '10:14',
                value: y
              },
              {
                name: 'Z',
                id: 'z',
                range: '14:18',
                value: z
              },
              {
                name: 'Yaw',
                id: 'yaw',
                range: '18:19',
                value: yaw
              },
              {
                name: 'Pitch',
                id: 'pitch',
                range: '19:20',
                value: pitch
              },
              {
                name: 'Head Yaw',
                id: 'headYaw',
                range: '20:21',
                value: headYaw
              },
            ],
            summary: 'Spawn Mob'
          };
        }
          break;
        */
        case 0xca:
        {
          let data = yield 4;
          let inv = !!data.readUInt8(0);
          let flying = !!data.readUInt8(1);
          let canFly = !!data.readUInt8(2);
          let destroy = !!data.readUInt8(3);
          yield {
            payload: Buffer.concat([idData, data]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'Invulnerability',
                id: 'inv',
                range: '1:2',
                value: inv
              },
              {
                name: 'Is flying',
                id: 'flying',
                range: '2:3',
                value: flying
              },
              {
                name: 'Can fly',
                id: 'canFly',
                range: '3:4',
                value: canFly
              },
              {
                name: 'Instant Destroy',
                id: 'destroy',
                range: '4:5',
                value: destroy
              },
            ],
            summary: 'Player Abilities'
          };
        }
          break;

        case 0xff:
        {
          let lenData = yield 2;
          let len = lenData.readUInt16BE(0);
          let msgData = (yield len * 2);
          let msg = utf16(msgData);
          yield {
            payload: Buffer.concat([idData, lenData, msgData]),
            items: [
              {
                name: 'Packet ID',
                id: 'id',
                range: '0:1',
                value: id
              },
              {
                name: 'Reason',
                id: 'reason',
                range: '2:',
                value: msg
              }
            ],
            summary: 'Disconnect/Kick'
          };
        }
          break;

        default:
          console.log('0x' + Buffer.from([id]).toString('hex'));
          yield null;
      }
    }
  }
};

// http://qiita.com/weal/items/10122402adb61597f851
function utf16(arr) {
  var endian = 'BE', i = 0, len, str;
  for(len=arr.length,str=""; i<len; i+=2) {
      str += String.fromCharCode(endian === 'LE' ? (arr[i+1]<<8|arr[i]) : (arr[i]<<8|arr[i+1]));
  }
  return str;
}
